import tensorflow as tf

from neupy.core.properties import (
    ProperFractionProperty,
    ParameterProperty,
    TypedListProperty,
    NumberProperty,
    IntProperty,
)
from neupy.utils import asfloat
from neupy.exceptions import (
    WeightInitializationError,
    LayerConnectionError,
)
from .base import Identity


__all__ = ('BatchNorm', 'LocalResponseNorm', 'GroupNorm')


class BatchNorm(Identity):
    """
    Batch normalization layer.

    Parameters
    ----------
    axes : tuple with ints or None
        Axes along which normalization will be applied. The ``None``
        value means that normalization will be applied over all axes
        except the last one. In case of 4D tensor it will
        be equal to ``(0, 1, 2)``. Defaults to ``None``.

    epsilon : float
        Epsilon is a positive constant that adds to the standard
        deviation to prevent the division by zero.
        Defaults to ``1e-5``.

    alpha : float
        Coefficient for the exponential moving average of
        batch-wise means and standard deviations computed during
        training; the closer to one, the more it will depend on
        the last batches seen. Value needs to be between ``0`` and ``1``.
        Defaults to ``0.1``.

    gamma : array-like, Tensorfow variable, scalar or Initializer
        Scale. Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=1)``.

    beta : array-like, Tensorfow variable, scalar or Initializer
        Offset. Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=0)``.

    running_mean : array-like, Tensorfow variable, scalar or Initializer
        Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=0)``.

    running_inv_std : array-like, Tensorfow variable, scalar or Initializer
        Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=1)``.

    {Identity.name}

    Methods
    -------
    {Identity.Methods}

    Attributes
    ----------
    {Identity.Attributes}

    Examples
    --------

    Feedforward Neural Networks (FNN) with batch normalization after
    activation function was applied.

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input(10),
    ...     Relu(5) >> BatchNorm(),
    ...     Relu(5) >> BatchNorm(),
    ...     Sigmoid(1),
    ... )

    Feedforward Neural Networks (FNN) with batch normalization before
    activation function was applied.

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input(10),
    ...     Linear(5) >> BatchNorm() >> Relu(),
    ...     Linear(5) >> BatchNorm() >> Relu(),
    ...     Sigmoid(1),
    ... )

    Convolutional Neural Networks (CNN)

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((28, 28, 1)),
    ...     Convolution((3, 3, 16)) >> BatchNorm() >> Relu(),
    ...     Convolution((3, 3, 16)) >> BatchNorm() >> Relu(),
    ...     Reshape(),
    ...     Softmax(10),
    ... )

    References
    ----------
    .. [1] Batch Normalization: Accelerating Deep Network Training
           by Reducing Internal Covariate Shift,
           http://arxiv.org/pdf/1502.03167v3.pdf
    """
    axes = TypedListProperty(allow_none=True)
    epsilon = NumberProperty(minval=0)
    alpha = ProperFractionProperty()
    beta = ParameterProperty()
    gamma = ParameterProperty()

    running_mean = ParameterProperty()
    running_inv_std = ParameterProperty()

    def __init__(self, axes=None, alpha=0.1, beta=0, gamma=1, epsilon=1e-5,
                 running_mean=0, running_inv_std=1, name=None):

        super(BatchNorm, self).__init__(name=name)

        self.axes = axes
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.epsilon = epsilon
        self.running_mean = running_mean
        self.running_inv_std = running_inv_std

        if axes is not None and len(set(axes)) != len(axes):
            raise ValueError(
                "Specified axes have to contain only unique values")

    def create_variables(self, input_shape):
        input_shape = tf.TensorShape(input_shape)

        if input_shape.ndims is None:
            raise WeightInitializationError(
                "Cannot initialize variables for the batch normalization "
                "layer, because input shape is undefined. Layer: {}"
                "".format(self))

        if self.axes is None:
            # If ndims == 4 then axes = (0, 1, 2)
            # If ndims == 2 then axes = (0,)
            self.axes = tuple(range(input_shape.ndims - 1))

        if any(axis >= input_shape.ndims for axis in self.axes):
            raise LayerConnectionError(
                "Batch normalization cannot be applied over one of "
                "the axis, because input has only {} dimensions. Layer: {}"
                "".format(input_shape.ndims, self))

        parameter_shape = tuple([
            input_shape[axis].value if axis not in self.axes else 1
            for axis in range(input_shape.ndims)
        ])

        if any(parameter is None for parameter in parameter_shape):
            unknown_dim_index = parameter_shape.index(None)

            raise WeightInitializationError(
                "Cannot create variables for batch normalization, because "
                "input has unknown dimension #{} (0-based indices). "
                "Input shape: {}, Layer: {}".format(
                    unknown_dim_index, input_shape, self))

        self.input_shape = input_shape
        self.running_mean = self.variable(
            value=self.running_mean, shape=parameter_shape,
            name='running_mean', trainable=False)

        self.running_inv_std = self.variable(
            value=self.running_inv_std, shape=parameter_shape,
            name='running_inv_std', trainable=False)

        self.gamma = self.variable(
            value=self.gamma, name='gamma',
            shape=parameter_shape)

        self.beta = self.variable(
            value=self.beta, name='beta',
            shape=parameter_shape)

    def output(self, input, training=False):
        input = tf.convert_to_tensor(input, dtype=tf.float32)

        if not training:
            mean = self.running_mean
            inv_std = self.running_inv_std
        else:
            alpha = asfloat(self.alpha)
            mean = tf.reduce_mean(
                input, self.axes,
                keepdims=True, name="mean",
            )
            variance = tf.reduce_mean(
                tf.squared_difference(input, tf.stop_gradient(mean)),
                self.axes,
                keepdims=True,
                name="variance",
            )
            inv_std = tf.rsqrt(variance + asfloat(self.epsilon))

            tf.add_to_collection(
                tf.GraphKeys.UPDATE_OPS,
                self.running_inv_std.assign(
                    asfloat(1 - alpha) * self.running_inv_std + alpha * inv_std
                )
            )
            tf.add_to_collection(
                tf.GraphKeys.UPDATE_OPS,
                self.running_mean.assign(
                    asfloat(1 - alpha) * self.running_mean + alpha * mean
                )
            )

        normalized_value = (input - mean) * inv_std
        return self.gamma * normalized_value + self.beta


class LocalResponseNorm(Identity):
    """
    Local Response Normalization Layer.

    Aggregation is purely across channels, not within channels,
    and performed "pixelwise".

    If the value of the :math:`i` th channel is :math:`x_i`, the output is

    .. math::
        x_i = \\frac{{x_i}}{{ (k + ( \\alpha \\sum_j x_j^2 ))^\\beta }}

    where the summation is performed over this position on :math:`n`
    neighboring channels.

    Parameters
    ----------
    alpha : float
        Coefficient, see equation above. Defaults to ``1e-4``.

    beta : float
        Offset, see equation above. Defaults to ``0.75``.

    k : float
        Exponent, see equation above. Defaults to ``2``.

    depth_radius : int
        Number of adjacent channels to normalize over, must be odd.
        Defaults to ``5``.

    {Identity.name}

    Methods
    -------
    {Identity.Methods}

    Attributes
    ----------
    {Identity.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = Input((10, 10, 12)) >> LocalResponseNorm()
    """
    alpha = NumberProperty()
    beta = NumberProperty()
    k = NumberProperty()
    depth_radius = IntProperty()

    def __init__(self, alpha=1e-4, beta=0.75, k=2, depth_radius=5, name=None):
        super(LocalResponseNorm, self).__init__(name=name)

        if depth_radius % 2 == 0:
            raise ValueError("Only works with odd `depth_radius` values")

        self.alpha = alpha
        self.beta = beta
        self.k = k
        self.depth_radius = depth_radius

    def get_output_shape(self, input_shape):
        if input_shape and input_shape.ndims != 4:
            raise LayerConnectionError(
                "Layer `{}` expected input with 4 dimensions, got {} instead. "
                "Shape: {}".format(self.name, input_shape.ndims, input_shape))

        return super(LocalResponseNorm, self).get_output_shape(input_shape)

    def output(self, input, **kwargs):
        return tf.nn.local_response_normalization(
            input,
            depth_radius=self.depth_radius,
            bias=self.k,
            alpha=self.alpha,
            beta=self.beta)


class GroupNorm(Identity):
    """
    Group Normalization layer. This layer is a simple alternative to the
    Batch Normalization layer for cases when batch size is small.

    Parameters
    ----------
    n_groups : int
        During normalization all the channels will be break down into
        separate groups and mean and variance will be estimated per group.
        This parameter controls number of groups.

    gamma : array-like, Tensorfow variable, scalar or Initializer
        Scale. Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=1)``.

    beta : array-like, Tensorfow variable, scalar or Initializer
        Offset. Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=0)``.

    epsilon : float
        Epsilon ensures that input rescaling procedure that uses estimated
        variance will never cause division by zero. Defaults to ``1e-5``.

    {Identity.name}

    Methods
    -------
    {Identity.Methods}

    Attributes
    ----------
    {Identity.Attributes}

    Examples
    --------
    Convolutional Neural Networks (CNN)

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((28, 28, 1)),
    ...     Convolution((3, 3, 16)) >> GroupNorm(4) >> Relu(),
    ...     Convolution((3, 3, 16)) >> GroupNorm(4) >> Relu(),
    ...     Reshape(),
    ...     Softmax(10),
    ... )

    References
    ----------
    .. [1] Group Normalization, Yuxin Wu, Kaiming He,
           https://arxiv.org/pdf/1803.08494.pdf
    """
    n_groups = IntProperty(minval=1)
    beta = ParameterProperty()
    gamma = ParameterProperty()
    epsilon = NumberProperty(minval=0)

    def __init__(self, n_groups, beta=0, gamma=1, epsilon=1e-5, name=None):
        super(GroupNorm, self).__init__(name=name)

        self.n_groups = n_groups
        self.beta = beta
        self.gamma = gamma
        self.epsilon = epsilon

    def create_variables(self, input_shape):
        n_channels = input_shape[3]

        if n_channels.value is None:
            raise WeightInitializationError(
                "Cannot initialize variables when number of "
                "channels is unknown. Input shape: {}, Layer: {}"
                "".format(input_shape, self))

        parameter_shape = (1, 1, 1, n_channels)

        self.gamma = self.variable(
            value=self.gamma, name='gamma',
            shape=parameter_shape)

        self.beta = self.variable(
            value=self.beta, name='beta',
            shape=parameter_shape)

    def get_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape)

        if input_shape and input_shape.ndims != 4:
            raise LayerConnectionError(
                "Group normalization layer expects 4 dimensional input, "
                "got {} instead. Input shape: {}, Layer: {}"
                "".format(input_shape.ndims, input_shape, self))

        n_channels = input_shape[3]

        if n_channels.value and n_channels % self.n_groups != 0:
            raise LayerConnectionError(
                "Cannot divide {} input channels into {} groups. "
                "Input shape: {}, Layer: {}".format(
                    n_channels, self.n_groups, input_shape, self))

        return super(GroupNorm, self).get_output_shape(input_shape)

    def output(self, input):
        input = tf.convert_to_tensor(input, dtype=tf.float32)
        input_shape = tf.shape(input)
        n_groups = self.n_groups

        # We access dimensional information in form of tensors in case
        # if some of the dimensions are undefined. In this way we make
        # sure that reshape will work even if part of the input shape
        # is undefined.
        dims = [input_shape[i] for i in range(4)]
        n_samples, height, width, n_channels = dims

        input = tf.reshape(input, [
            n_samples, height, width, n_groups, n_channels // n_groups])

        mean, variance = tf.nn.moments(input, [1, 2, 4], keep_dims=True)
        input = (input - mean) / tf.sqrt(variance + self.epsilon)
        input = tf.reshape(input, input_shape)

        return input * self.gamma + self.beta



from __future__ import division

import tensorflow as tf

from neupy.core.properties import ProperFractionProperty
from neupy.utils import flatten, make_single_vector
from neupy.utils.tf_utils import setup_parameter_updates

from .base import BaseOptimizer


__all__ = ('HessianDiagonal',)


class HessianDiagonal(BaseOptimizer):
    """
    Algorithm that uses calculates only diagonal values from the Hessian matrix
    and uses it instead of the Hessian matrix.

    Parameters
    ----------
    min_eigval : float
        Set up minimum eigenvalue for Hessian diagonal matrix. After a few
        iteration elements will be extremely small and matrix inverse
        produce huge number in hessian diagonal elements. This
        parameter control diagonal elements size. Defaults to ``1e-2``.

    {BaseOptimizer.Parameters}

    Attributes
    ----------
    {BaseOptimizer.Attributes}

    Methods
    -------
    {BaseOptimizer.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.HessianDiagonal(network)
    >>> optimizer.train(x_train, y_train)

    Notes
    -----
    - Method requires all training data during propagation, which means
      it cannot be trained with mini-batches.

    See Also
    --------
    :network:`BaseOptimizer` : BaseOptimizer algorithm.
    :network:`Hessian` : Newton's method.
    """
    min_eigval = ProperFractionProperty(default=1e-2)

    def init_train_updates(self):
        step = self.step
        inv_min_eigval = 1 / self.min_eigval
        variables = self.network.variables
        parameters = [var for var in variables.values() if var.trainable]
        param_vector = make_single_vector(parameters)

        gradients = tf.gradients(self.variables.loss, parameters)
        full_gradient = make_single_vector(gradients)

        second_derivatives = []
        for parameter, gradient in zip(parameters, gradients):
            second_derivative, = tf.gradients(gradient, parameter)
            second_derivatives.append(flatten(second_derivative))

        hessian_diag = tf.concat(second_derivatives, axis=0)

        # it's easier to clip inverse hessian rather than the hessian,.
        inv_hessian_diag = tf.clip_by_value(
            # inverse for diagonal matrix easy to compute with
            # elementwise inverse operation.
            1 / hessian_diag,
            -inv_min_eigval,
            inv_min_eigval,
        )
        updates = setup_parameter_updates(
            parameters,
            param_vector - step * full_gradient * inv_hessian_diag
        )
        return updates



import re
import copy
import types
import string
import inspect
from functools import partial
from collections import OrderedDict, defaultdict

import six
import numpy as np
import tensorflow as tf

from neupy.exceptions import LayerConnectionError
from neupy.core.properties import Property, TypedListProperty
from neupy.utils import as_tuple, tf_utils
from neupy.layers.graph import BaseGraph, make_one_if_possible


__all__ = ('BaseLayer', 'Identity', 'Input')


def create_name_pattern_from_layer_name(layer):
    classname = layer.__class__.__name__
    layer_name = re.sub(r'(?<!^)(?=[A-Z][a-z_])', '-', classname)
    return layer_name.lower() + "-{}"


def format_name_if_specified_as_pattern(name):
    if not hasattr(format_name_if_specified_as_pattern, 'counters'):
        format_name_if_specified_as_pattern.counters = defaultdict(lambda: 1)

    formatter = string.Formatter()
    variables = [val[1] for val in formatter.parse(name) if val[1] is not None]

    if len(variables) == 0:
        return name

    if len(variables) >= 2:
        raise ValueError(
            "Provided pattern has more than one field specified. "
            "Pattern: {}".format(name))

    layer_id = format_name_if_specified_as_pattern.counters[name]
    formatted_name = name.format(layer_id)

    # Increment only after string was successfully formatted
    format_name_if_specified_as_pattern.counters[name] += 1
    return formatted_name


class BaseLayer(BaseGraph):
    """
    Base class for the layers.

    Parameters
    ----------
    name : str or None
        Layer's name. Can be used as a reference to specific layer. Name
        Can be specified as:

            - String: Specified name will be used as a direct reference to
              the layer. For example, `name="fc"`

            - Format string: Name pattern could be defined as a format string
              and specified field will be replaced with an index. For example,
              `name="fc{{}}"` will be replaced with `fc1`, `fc2` and so on.
              A bit more complex formatting methods are acceptable, for
              example, `name="fc-{{:<03d}}"` will be converted to `fc-001`,
              `fc-002`, `fc-003` and so on.

            - ``None``: When value specified as ``None`` than name will be
              generated from the class name.

        Defaults to ``None``.

    Methods
    -------
    variable(value, name, shape=None, trainable=True)
        Initializes variable with specified values.

    get_output_shape(input_shape)
        Computes expected output shape from the layer based on the
        specified input shape.

    output(*inputs, **kwargs)
        Propagates input through the layer. The ``kwargs``  variable
        might contain additional information that propagates through the
        network.

    Attributes
    ----------
    variables : dict
        Variable names and their values. Dictionary can be empty in case
        if variables hasn't been created yet.
    """
    name = Property(expected_type=six.string_types)

    def __init__(self, name=None):
        # Layer by default gets intialized as a graph with single node in it
        super(BaseLayer, self).__init__(forward_graph=[(self, [])])
        self.original_name = name

        if name is None:
            name = create_name_pattern_from_layer_name(layer=self)

        name = format_name_if_specified_as_pattern(name)

        self.variables = OrderedDict()
        self.name = name

        self._input_shape = tf.TensorShape(None)
        self.frozen = False

        # This decorator ensures that result produced by the
        # `output` method will be marked under layer's name scope.
        self.output = types.MethodType(
            tf_utils.class_method_name_scope(self.output), self)

    @classmethod
    def define(cls, *args, **kwargs):
        return partial(cls, *args, **kwargs)

    @property
    def input_shape(self):
        # Explicit TensorShape transformation not only ensures
        # that we have right type in the output, but also copies
        # value stored in the `_input_shape` in order to make sure
        # that no in-place update can effect original value
        return tf.TensorShape(self._input_shape)

    @input_shape.setter
    def input_shape(self, shape):
        if not self._input_shape.is_compatible_with(shape):
            raise ValueError(
                "Cannot update input shape of the layer, because it's "
                "incompatible with current input shape. Current shape: {}, "
                "New shape: {}, Layer: {}".format(
                    self._input_shape, shape, self))

        self._input_shape = tf.TensorShape(shape)

    @property
    def output_shape(self):
        return self.get_output_shape(self.input_shape)

    def get_output_shape(self, input_shape):
        return tf.TensorShape(None)

    def create_variables(self, *input_shapes):
        return NotImplemented

    def variable(self, value, name, shape=None, trainable=True):
        layer_name = 'layer/{layer_name}/{parameter_name}'.format(
            layer_name=self.name,
            parameter_name=name.replace('_', '-'))

        self.variables[name] = tf_utils.create_variable(
            value, layer_name, shape, trainable)

        return self.variables[name]

    def _repr_arguments(self, *args, **kwargs):
        def format_value(value):
            references = {
                'Variable': tf.Variable,
                'Array': np.ndarray,
                'Matrix': np.matrix,
            }

            for name, datatype in references.items():
                if isinstance(value, datatype):
                    return '<{} shape={}>'.format(name, value.shape)

            return repr(value)

        formatted_args = [str(arg) for arg in args]
        argspec = inspect.getargspec(self.__class__.__init__)

        def kwargs_priority(value):
            if value in argspec.args:
                return argspec.args.index(value)
            return float('inf')

        # Kwargs will have destroyed order of the arguments, and order in
        # the __init__ method allows to use proper order and validate names
        for name in sorted(kwargs.keys(), key=kwargs_priority):
            value = format_value(kwargs[name])
            formatted_args.append('{}={}'.format(name, value))

        return '{clsname}({formatted_args})'.format(
            clsname=self.__class__.__name__,
            formatted_args=', '.join(formatted_args))

    def __copy__(self):
        params = self.get_params()
        # We make sure that new name will be created if user
        # specified name=None
        params['name'] = self.original_name

        copied_layer = self.__class__(**params)
        # Input shape can change and it cannot be controlled from
        # the __init__ method
        copied_layer.input_shape = self.input_shape

        return copied_layer

    def __deepcopy__(self, memo):
        memo[id(self)] = copied_layer = self.__copy__()

        for key, new_value in copied_layer.get_params().items():
            old_value = getattr(self, key)

            # Some of the objects could be already copied
            # (for example, tensorflow's variables)
            if old_value is new_value:
                copied_new_value = copy.deepcopy(new_value, memo)
                setattr(copied_layer, key, copied_new_value)

        return copied_layer

    def __repr__(self):
        kwargs = {}

        for name in self.options:
            value = getattr(self, name)
            kwargs[name] = value

        return self._repr_arguments(**kwargs)


class Identity(BaseLayer):
    """
    Passes input through the layer without changes. Can be
    useful while defining residual networks in the network.

    Parameters
    ----------
    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    def get_output_shape(self, input_shape):
        return tf.TensorShape(input_shape)

    def output(self, input, **kwargs):
        return input


class Input(BaseLayer):
    """
    Layer defines network's input.

    Parameters
    ----------
    shape : int or tuple
        Shape of the input features per sample. Batch
        dimension has to be excluded from the shape.

    {BaseLayer.name}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    Feedforward Neural Network (FNN)

    In the example, input layer defines network that expects
    2D inputs (matrices). In other words, input to the network
    should be set of samples combined into matrix where each sample
    has 10 dimensional vector associated with it.

    >>> from neupy.layers import *
    >>> network = Input(10) >> Relu(5) >> Softmax(3)

    Convolutional Neural Network (CNN)

    In the example, input layer specified that we expect multiple
    28x28 image as an input and each image should have single
    channel (images with no color).

    >>> from neupy.layers import *
    >>> network = join(
    ...     Input((28, 28, 1)),
    ...     Convolution((3, 3, 16)) >> Relu(),
    ...     Convolution((3, 3, 16)) >> Relu(),
    ...     Reshape()
    ...     Softmax(10),
    ... )
    """
    shape = TypedListProperty(element_type=(int, type(None)))

    def __init__(self, shape, name=None):
        super(Input, self).__init__(name=name)

        if isinstance(shape, tf.TensorShape):
            shape = tf_utils.shape_to_tuple(shape)

        self.shape = as_tuple(shape)

    @BaseLayer.input_shape.getter
    def input_shape(self):
        batch_shape = tf.TensorShape([None])
        return batch_shape.concatenate(self.shape)

    def output(self, input, **kwargs):
        return input

    def get_output_shape(self, input_shape):
        if not self.input_shape.is_compatible_with(input_shape):
            raise LayerConnectionError(
                "Input layer got unexpected input shape. "
                "Received shape: {}, Expected shape: {}"
                "".format(input_shape, self.input_shape)
            )
        return self.input_shape.merge_with(input_shape)

    def __repr__(self):
        return self._repr_arguments(
            make_one_if_possible(self.shape),
            name=self.name)



import tensorflow as tf
import numpy as np

from neupy.utils import tensorflow_session
from neupy.core.properties import BoundedProperty, ProperFractionProperty
from .base import BaseOptimizer


__all__ = ('RPROP', 'IRPROPPlus')


class RPROP(BaseOptimizer):
    """
    Resilient backpropagation (RPROP) is an optimization
    algorithm for supervised learning.

    RPROP algorithm takes into account only direction of the gradient
    and completely ignores its magnitude. Every weight values has a unique
    step size associated with it (by default all of the are equal to ``step``).

    The rule is following, when gradient direction changes (sign of the
    gradient) we decrease step size for specific weight multiplying it by
    ``decrease_factor`` and if sign stays the same than we increase step
    size for this specific weight multiplying it by ``increase_factor``.

    The step size is always bounded by ``minstep`` and ``maxstep``.

    Notes
    -----
    Algorithm doesn't work with mini-batches.

    Parameters
    ----------
    minstep : float
        Minimum possible value for step. Defaults to ``0.001``.

    maxstep : float
        Maximum possible value for step. Defaults to ``10``.

    increase_factor : float
        Increase factor for step in case when gradient doesn't change
        sign compare to previous epoch.

    decrease_factor : float
        Decrease factor for step in case when gradient changes sign
        compare to previous epoch.

    {BaseOptimizer.Parameters}

    Attributes
    ----------
    {BaseOptimizer.Attributes}

    Methods
    -------
    {BaseOptimizer.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.RPROP(network)
    >>> optimizer.train(x_train, y_train)

    See Also
    --------
    :network:`IRPROPPlus` : iRPROP+ algorithm.
    :network:`GradientDescent` : GradientDescent algorithm.
    """

    # This properties correct upper and lower bounds for steps.
    minstep = BoundedProperty(default=0.001, minval=0)
    maxstep = BoundedProperty(default=10, minval=0)

    # This properties increase/decrease step by dividing it to
    # some coefficient.
    increase_factor = BoundedProperty(minval=1, default=1.2)
    decrease_factor = ProperFractionProperty(default=0.5)

    def update_prev_delta(self, prev_delta):
        return prev_delta

    def init_train_updates(self):
        updates = []

        variables = []
        for (_, _), variable in self.network.variables.items():
            if variable.trainable:
                variables.append(variable)

        gradients = tf.gradients(self.variables.loss, variables)

        for parameter, gradient in zip(variables, gradients):
            with tf.variable_scope(parameter.op.name):
                steps = tf.Variable(
                    # Steps will be decreased after the first iteration,
                    # because all previous gradients are equal to zero.
                    # In order to make sure that network will use the same
                    # step per every weight we re-scale step and after the
                    # first iteration it will be multiplied by
                    # ``decrease_factor`` and scaled back to the default
                    # step value.
                    tf.ones_like(parameter) * self.step,
                    name="steps",
                    dtype=tf.float32,
                )
                prev_delta = tf.Variable(
                    tf.zeros(parameter.shape),
                    name="prev-delta",
                    dtype=tf.float32,
                )
                # We collect only signs since it ensures numerical stability
                # after multiplication when we deal with small numbers.
                prev_gradient_sign = tf.Variable(
                    tf.zeros(parameter.shape),
                    name="prev-grad-sign",
                    dtype=tf.float32,
                )

            updated_prev_delta = self.update_prev_delta(prev_delta)
            gradient_sign = tf.sign(gradient)

            grad_sign_product = gradient_sign * prev_gradient_sign
            gradient_changed_sign = tf.equal(grad_sign_product, -1)

            updated_steps = tf.clip_by_value(
                tf.where(
                    tf.equal(grad_sign_product, 1),
                    steps * self.increase_factor,
                    tf.where(
                        gradient_changed_sign,
                        steps * self.decrease_factor,
                        steps,
                    )
                ),
                self.minstep,
                self.maxstep,
            )
            parameter_delta = tf.where(
                gradient_changed_sign,
                # If we subtract previous negative weight update it means
                # that we will revert weight update that has been  applied
                # in the previous iteration.
                -updated_prev_delta,
                updated_steps * gradient_sign,
            )
            # Making sure that during the next iteration sign, after
            # we multiplied by the new gradient, won't be negative.
            # Otherwise, the same roll back using previous delta
            # won't make much sense.
            clipped_gradient_sign = tf.where(
                gradient_changed_sign,
                tf.zeros_like(gradient_sign),
                gradient_sign,
            )

            updates.extend([
                (parameter, parameter - parameter_delta),
                (steps, updated_steps),
                (prev_gradient_sign, clipped_gradient_sign),
                (prev_delta, parameter_delta),
            ])

        return updates


class IRPROPPlus(RPROP):
    """
    iRPROP+ is an optimization algorithm for supervised learning.
    This is a variation of the :network:`RPROP` algorithm.

    Parameters
    ----------
    {RPROP.minstep}

    {RPROP.maxstep}

    {RPROP.increase_factor}

    {RPROP.decrease_factor}

    {BaseOptimizer.regularizer}

    {BaseOptimizer.network}

    {BaseOptimizer.loss}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.signals}

    {Verbose.verbose}

    Methods
    -------
    {BaseSkeleton.predict}

    {BaseOptimizer.train}

    {BaseSkeleton.fit}

    Notes
    -----
    {RPROP.Notes}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> network = Input(2) >> Sigmoid(3) >> Sigmoid(1)
    >>> optimizer = algorithms.IRPROPPlus(network)
    >>> optimizer.train(x_train, y_train)

    References
    ----------
    [1] Christian Igel, Michael Huesken (2000)
        Improving the Rprop Learning Algorithm

    See Also
    --------
    :network:`RPROP` : RPROP algorithm.
    :network:`GradientDescent` : GradientDescent algorithm.
    """
    def init_functions(self):
        self.variables.update(
            last_error=tf.Variable(np.nan, name='irprop-plus/last-error'),
            previous_error=tf.Variable(
                np.nan, name='irprop-plus/previous-error'),
        )
        super(IRPROPPlus, self).init_functions()

    def one_training_update(self, X_train, y_train):
        if len(self.errors.train) >= 2:
            previous_error, last_error = self.errors.train[-2:]
            session = tensorflow_session()

            self.variables.last_error.load(last_error, session)
            self.variables.previous_error.load(previous_error, session)

        return super(IRPROPPlus, self).one_training_update(X_train, y_train)

    def update_prev_delta(self, prev_delta):
        last_error = self.variables.last_error
        prev_error = self.variables.previous_error

        return tf.where(
            # We revert weight when gradient changed the sign only in
            # cases when error increased. Otherwise we don't apply any
            # update for this weight.
            last_error > prev_error,
            prev_delta,
            tf.zeros_like(prev_delta),
        )



import numpy as np

from neupy import init
from neupy.utils import format_data
from neupy.core.properties import IntProperty, ParameterProperty, ArrayProperty
from neupy.algorithms.base import BaseNetwork


__all__ = ('BaseStepAssociative',)


class BaseAssociative(BaseNetwork):
    """
    Base class for associative learning.

    Parameters
    ----------
    n_inputs : int
        Number of features (columns) in the input data.

    n_outputs : int
        Number of outputs in the  network.

    weight : array-like, Initializer
        Neural network weights.
        Value defined manually should have shape ``(n_inputs, n_outputs)``.
        Defaults to :class:`Normal() <neupy.init.Normal>`.

    {BaseNetwork.Parameters}

    Methods
    -------
    {BaseSkeleton.predict}

    train(X_train, epochs=100)
        Train neural network.

    {BaseSkeleton.fit}
    """
    n_inputs = IntProperty(minval=1, required=True)
    n_outputs = IntProperty(minval=1, required=True)
    weight = ParameterProperty(default=init.Normal())

    def __init__(self, **options):
        super(BaseAssociative, self).__init__(**options)
        self.init_weights()

    def init_weights(self):
        valid_weight_shape = (self.n_inputs, self.n_outputs)

        if isinstance(self.weight, init.Initializer):
            self.weight = self.weight.sample(
                valid_weight_shape, return_array=True)

        if self.weight.shape != valid_weight_shape:
            raise ValueError(
                "Weight matrix has invalid shape. Got {}, expected {}"
                "".format(self.weight.shape, valid_weight_shape))

        self.weight = self.weight.astype(float)

    def format_input_data(self, X):
        X = format_data(X, is_feature1d=(self.n_inputs == 1))

        if X.ndim != 2:
            raise ValueError(
                "Cannot make prediction, because input "
                "data has more than 2 dimensions")

        if X.shape[1] != self.n_inputs:
            raise ValueError(
                "Input data expected to have {} features, "
                "but got {}".format(self.n_inputs, X.shape[1]))

        return X

    def train(self, X_train, epochs=100):
        X_train = self.format_input_data(X_train)
        return super(BaseAssociative, self).train(
            X_train=X_train, epochs=epochs)


class BaseStepAssociative(BaseAssociative):
    """
    Base class for associative algorithms which have 2 layers and first
    one is has step function as activation.

    Parameters
    ----------
    {BaseAssociative.n_inputs}

    {BaseAssociative.n_outputs}

    n_unconditioned : int
        Number of unconditioned units in neural networks. All these
        units wouldn't update during the training procedure.
        Unconditioned should be the first feature in the dataset.

    weight : array-like
        Neural network weights.
        Value defined manually should have shape ``(n_inputs, n_outputs)``.
        Defaults to ``None`` which means that all unconditional
        weights will be equal to ``1``. Other weights equal to ``0``.

    bias : array-like, Initializer
        Neural network bias units.
        Defaults to :class:`Constant(-0.5) <neupy.init.Constant>`.

    {BaseNetwork.Parameters}

    Methods
    -------
    {BaseAssociative.Methods}
    """
    n_inputs = IntProperty(minval=2, required=True)
    n_unconditioned = IntProperty(minval=1, required=True)

    weight = ArrayProperty()
    bias = ParameterProperty(default=init.Constant(-0.5))

    def init_weights(self):
        if self.n_inputs <= self.n_unconditioned:
            raise ValueError(
                "Number of unconditioned features should be less than total "
                "number of features. `n_inputs`={} and `n_unconditioned`={}"
                "".format(self.n_inputs, self.n_unconditioned))

        valid_weight_shape = (self.n_inputs, self.n_outputs)
        valid_bias_shape = (self.n_outputs,)

        if self.weight is None:
            self.weight = np.zeros(valid_weight_shape)
            self.weight[:self.n_unconditioned, :] = 1

        if isinstance(self.bias, init.Initializer):
            self.bias = self.bias.sample(valid_bias_shape, return_array=True)

        super(BaseStepAssociative, self).init_weights()

        if self.bias.shape != valid_bias_shape:
            raise ValueError(
                "Bias vector has invalid shape. Got {}, expected {}"
                "".format(self.bias.shape, valid_bias_shape))

        self.bias = self.bias.astype(float)

    def predict(self, X):
        X = format_data(X, is_feature1d=False)
        raw_output = X.dot(self.weight) + self.bias
        return np.where(raw_output > 0, 1, 0)

    def train(self, X_train, *args, **kwargs):
        X_train = format_data(X_train, is_feature1d=False)
        return super(BaseStepAssociative, self).train(X_train, *args, **kwargs)

    def one_training_update(self, X_train, y_train):
        weight = self.weight
        n_unconditioned = self.n_unconditioned
        predict = self.predict
        weight_delta = self.weight_delta

        error = 0

        for x_row in X_train:
            x_row = np.expand_dims(x_row, axis=0)
            layer_output = predict(x_row)

            delta = weight_delta(x_row, layer_output)
            weight[n_unconditioned:, :] += delta

            # This error can tell us whether network has converged
            # to some value of weights. Low errors will mean that weights
            # hasn't been updated much during the training epoch.
            error += np.linalg.norm(delta)

        return error



import tensorflow as tf

from neupy.utils import asfloat, function_name_scope


__all__ = ('step_decay', 'exponential_decay', 'polynomial_decay')


def init_variables(initial_value, iteration=0, name='step'):
    iteration = tf.Variable(
        asfloat(iteration),
        dtype=tf.float32,
        name='iteration',
    )
    step = tf.Variable(
        asfloat(initial_value),
        dtype=tf.float32,
        name=name,
    )
    return step, iteration


@function_name_scope
def step_decay(initial_value, reduction_freq, start_iter=0, name='step'):
    """
    Algorithm minimizes learning step monotonically after
    each iteration.

    .. math::
        \\alpha_{t + 1} = \\frac{\\alpha_{0}}{1 + \\frac{t}{m}}

    where :math:`\\alpha` is a step, :math:`t` is an iteration number
    and :math:`m` is a ``reduction_freq`` parameter.

    .. code-block:: python

        step = initial_value / (1 + current_iteration / reduction_freq)

    Notes
    -----
    Step will be reduced faster when you have smaller training batches.

    Parameters
    ----------
    initial_value : float
        Initial value for the learning rate. It's the learning rate
        returned during the first iteration.

    reduction_freq : int
        Parameter controls step reduction frequency. The larger the
        value the slower step parameter decreases.

        For instance, if ``reduction_freq=100``
        and ``step=0.12`` then after ``100`` iterations ``step`` is
        going to be equal to ``0.06`` (which is ``0.12 / 2``),
        after ``200`` iterations ``step`` is going to be equal to
        ``0.04`` (which is ``0.12 / 3``) and so on.

    start_iter : int
        Start iteration. At has to be equal to ``0`` when network just
        started the training. Defaults to ``0``.

    name : str
        Learning rate's variable name. Defaults to ``step``.

    Examples
    --------
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.step_decay(
    ...         initial_value=0.1,
    ...         reduction_freq=100,
    ...     )
    ... )
    """
    step, iteration = init_variables(initial_value, start_iter, name)
    reduction_freq = asfloat(reduction_freq)

    step_update = initial_value / (1 + iteration / reduction_freq)
    updated_step = step.assign(step_update)
    tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, updated_step)

    with tf.control_dependencies([updated_step]):
        next_iteration = iteration.assign(iteration + 1)
        tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, next_iteration)

    return step


@function_name_scope
def exponential_decay(initial_value, reduction_freq, reduction_rate,
                      staircase=False, start_iter=0, name='step'):
    """
    Applies exponential decay to the learning rate. This function is a
    wrapper for the tensorflow's ``exponential_decay`` function.

    .. math::
        \\alpha_{t + 1} = \\alpha_{0} \\cdot d^{\\frac{t}{r}}

    where :math:`\\alpha` is a step, :math:`t` is an iteration number,
    :math:`d` is a ``reduction_freq`` and :math:`r` is a ``reduction_rate``.

    .. code-block:: python

        step = initial_value * reduction_rate ^ (
            current_iteration / reduction_freq)

    When ``staircase=True`` the :math:`\\frac{t}{r}` value will be
    rounded.

    .. code-block:: python

        step = initial_value * reduction_rate ^ floor(
            current_iteration / reduction_freq)

    Notes
    -----
    Step will be reduced faster when you have smaller training batches.

    Parameters
    ----------
    initial_value : float
        Initial value for the learning rate.

    reduction_freq : int
        Parameter controls step reduction frequency. The larger the
        value the slower step parameter decreases.

    reduction_rate : float
        Parameter controls step reduction rate. The larger the
        value the slower step parameter decreases.

    staircase : bool
         If ``True`` decay the learning rate at discrete intervals.
         Defaults to ``False``.

    start_iter : int
        Start iteration. At has to be equal to ``0`` when network just
        started the training. Defaults to ``0``.

    name : str
        Learning rate's variable name. Defaults to ``step``.

    Examples
    --------
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.exponential_decay(
    ...         initial_value=0.1,
    ...         reduction_freq=1000,
    ...         reduction_rate=0.95,
    ...     )
    ... )
    """
    step, iteration = init_variables(initial_value, start_iter, name)
    step_update = tf.train.exponential_decay(
        learning_rate=initial_value,
        global_step=iteration,
        decay_steps=reduction_freq,
        decay_rate=reduction_rate,
        staircase=staircase,
    )

    updated_step = step.assign(step_update)
    tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, updated_step)

    with tf.control_dependencies([updated_step]):
        next_iteration = iteration.assign(iteration + 1)
        tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, next_iteration)

    return step


@function_name_scope
def polynomial_decay(initial_value, decay_iter, minstep=0.001, power=1.0,
                     cycle=False, start_iter=0, name='step'):
    """
    Applies polynomial decay to the learning rate. This function is a
    wrapper for the tensorflow's ``polynomial_decay`` function.

    .. code-block:: python

        iteration = min(current_iteration, decay_iter)
        step = minstep + (
            (initial_value - minstep) *
            (1 - iteration / decay_iter) ^ power
        )

    If cycle is ``True`` then a multiple of ``decay_iter`` is used,
    the first one that is bigger than ``current_iterations``.

    .. code-block:: python

        decay_iter = decay_iter * ceil(current_iteration / decay_iter)
        step = minstep + (
            (initial_value - minstep) *
            (1 - current_iteration / decay_iter) ^ power
        )

    Notes
    -----
    Step will be reduced faster when you have smaller training batches.

    Parameters
    ----------
    initial_value : float
       Initial value for the learning rate.

    decay_iter : int
        When ``cycle=False`` parameter identifies number of iterations
        when ``minstep`` will be reached. When ``cycle=True`` than
        the ``decay_iter`` value will be increased. See code above.

    minstep : float
        Step will never be lower than that minimum possible step,
        specified by this parameter. Defaults to ``0.001``.

    power : float
        The power of the polynomial. Defaults to ``1``.

    cycle : bool
        When value equal to ``True`` than step will be further reduced
        when ``current_iteration > decay_iter``. Defaults to ``False``.

    start_iter : int
        Start iteration. At has to be equal to ``0`` when network just
        started the training. Defaults to ``0``.

    name : str
       Learning rate's variable name. Defaults to ``step``.

    Examples
    --------
    >>> from neupy import algorithms
    >>> from neupy.layers import *
    >>>
    >>> optimizer = algorithms.Momentum(
    ...     Input(5) >> Relu(10) >> Sigmoid(1),
    ...     step=algorithms.polynomial_decay(
    ...         initial_value=0.1,
    ...         decay_iter=1000,
    ...         minstep=0.01,
    ...     )
    ... )
    """
    step, iteration = init_variables(initial_value, start_iter, name)
    step_update = tf.train.polynomial_decay(
        learning_rate=initial_value,
        global_step=iteration,
        decay_steps=decay_iter,
        end_learning_rate=minstep,
        power=power,
        cycle=cycle,
    )

    updated_step = step.assign(step_update)
    tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, updated_step)

    with tf.control_dependencies([updated_step]):
        next_iteration = iteration.assign(iteration + 1)
        tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, next_iteration)

    return step



from neupy import layers
from neupy.utils import function_name_scope


__all__ = ('resnet50',)


@function_name_scope
def ResidualUnit(n_input_filters, stride=1, rate=1, has_branch=False,
                 name=None):

    def bn_name(index):
        return 'bn' + name + '_branch' + index

    def conv_name(index):
        return 'res' + name + '_branch' + index

    n_output_filters = 4 * n_input_filters
    main_branch = layers.join(
        # The main purpose of this 1x1 convolution layer is to
        # reduce number of filters. For instance, for the tensor with
        # 256 filters it can be reduced to 64. This trick allows to
        # reduce computation by factor of 4.
        layers.Convolution(
            size=(1, 1, n_input_filters),
            stride=stride,
            bias=None,
            name=conv_name('2a'),
        ),
        layers.BatchNorm(name=bn_name('2a')),
        layers.Relu(),

        # This convolution layer applies 3x3 filter in order to
        # extract features.
        layers.Convolution(
            (3, 3, n_input_filters),
            padding='same',
            dilation=rate,
            bias=None,
            name=conv_name('2b'),
        ),
        layers.BatchNorm(name=bn_name('2b')),
        layers.Relu(),

        # Last layer reverses operations of the first layer. In this
        # case we increase number of filters. For instance, from previously
        # obtained 64 filters we can increase it back to the 256 filters
        layers.Convolution(
            (1, 1, n_output_filters),
            bias=None,
            name=conv_name('2c')
        ),
        layers.BatchNorm(name=bn_name('2c')),
    )

    if has_branch:
        residual_branch = layers.join(
            layers.Convolution(
                (1, 1, n_output_filters),
                stride=stride,
                bias=None,
                name=conv_name('1'),
            ),
            layers.BatchNorm(name=bn_name('1')),
        )
    else:
        # Empty list defines residual connection, meaning that
        # output from this branch would be equal to its input
        residual_branch = layers.Identity('residual-' + name)

    return layers.join(
        # For the output from two branches we just combine results
        # with simple elementwise sum operation. The main purpose of
        # the residual connection is to build shortcuts for the
        # gradient during backpropagation.
        (main_branch | residual_branch),
        layers.Elementwise(),
        layers.Relu(),
    )


def resnet50(input_shape=(224, 224, 3), include_global_pool=True,
             in_out_ratio=32):
    """
    ResNet50 network architecture with random parameters. Parameters
    can be loaded using ``neupy.storage`` module.

    ResNet50 has roughly 25.5 million parameters.

    Parameters
    ----------
    input_shape : tuple
        Network's input shape. Defaults to ``(224, 224, 3)``.

    include_global_pool : bool
        Specifies if returned output should include global pooling
        layer. Defaults to ``True``.

    in_out_ratio : {4, 8, 16, 32}
        Every layer that applies strides reduces height and width per every
        image. There are 5 of these layers in Resnet and at the end each
        dimensions gets reduced by ``32``. For example, 224x224 image
        will be reduced to 7x7 image patches. This parameter specifies
        what level of reduction we want to obtain after we've propagated
        network through all the convolution layers.

    Notes
    -----
    Because of the global pooling layer, ResNet50 can be applied to
    the images with variable sizes. The only limitation is that image
    size should be bigger than 32x32, otherwise network won't be able
    to apply all transformations to the image.

    Examples
    --------
    ResNet-50 for ImageNet classification

    >>> from neupy import architectures, algorithms
    >>>
    >>> resnet = architectures.resnet50()
    >>> resnet
    (?, 224, 224, 3) -> [... 187 layers ...] -> (?, 1000)
    >>>
    >>> optimizer = algorithms.Momentum(resnet50)

    ResNet-50 for custom classification task

    >>> from neupy import architectures
    >>> resnet = architectures.resnet50(include_global_pool=False)
    >>> resnet
    (?, 224, 224, 3) -> [... 185 layers ...] -> (?, 7, 7, 2048)
    >>>
    >>> from neupy.layers import *
    >>> resnet = resnet >> GlobalPooling('avg') >> Softmax(21)
    (?, 224, 224, 3) -> [... 187 layers ...] -> (?, 21)

    ResNet-50 for image segmentation

    >>> from neupy import architectures
    >>> resnet = architectures.resnet50(
    ...     include_global_pool=False,
    ...     in_out_ratio=8,
    ... )
    >>> resnet
    (?, 224, 224, 3) -> [... 185 layers ...] -> (?, 28, 28, 2048)

    See Also
    --------
    :architecture:`vgg16` : VGG16 network
    :architecture:`squeezenet` : SqueezeNet network
    :architecture:`resnet50` : ResNet-50 network

    References
    ----------
    Deep Residual Learning for Image Recognition.
    https://arxiv.org/abs/1512.03385
    """
    in_out_configs = {
        4: {'strides': [1, 1, 1], 'rates': [2, 4, 8]},
        8: {'strides': [2, 1, 1], 'rates': [1, 2, 4]},
        16: {'strides': [2, 2, 1], 'rates': [1, 1, 2]},
        32: {'strides': [2, 2, 2], 'rates': [1, 1, 1]},
    }

    if in_out_ratio not in in_out_configs:
        raise ValueError(
            "Expected one of the folowing in_out_ratio values: {}, got "
            "{} instead.".format(in_out_configs.keys(), in_out_ratio))

    strides = in_out_configs[in_out_ratio]['strides']
    rates = in_out_configs[in_out_ratio]['rates']

    resnet = layers.join(
        layers.Input(input_shape),

        # Convolutional layer reduces image's height and width by a factor
        # of 2 (because of the stride)
        # from (3, 224, 224) to (64, 112, 112)
        layers.Convolution(
            (7, 7, 64), stride=2, bias=None,
            padding='same', name='conv1'
        ),
        layers.BatchNorm(name='bn_conv1'),
        layers.Relu(),

        # Stride equal two 2 reduces image size by a factor of two
        # from (64, 112, 112) to (64, 56, 56)
        layers.MaxPooling((3, 3), stride=2, padding="same"),

        # The branch option applies extra convolution x+ batch
        # normalization transformations to the residual
        ResidualUnit(64, name='2a', has_branch=True),
        ResidualUnit(64, name='2b'),
        ResidualUnit(64, name='2c'),

        # When stride=2 reduces width and hight by factor of 2
        ResidualUnit(128, stride=strides[0], name='3a', has_branch=True),
        ResidualUnit(128, rate=rates[0], name='3b'),
        ResidualUnit(128, rate=rates[0], name='3c'),
        ResidualUnit(128, rate=rates[0], name='3d'),

        # When stride=2 reduces width and hight by factor of 2
        ResidualUnit(256, rate=rates[0], name='4a',
                     stride=strides[1], has_branch=True),
        ResidualUnit(256, rate=rates[1], name='4b'),
        ResidualUnit(256, rate=rates[1], name='4c'),
        ResidualUnit(256, rate=rates[1], name='4d'),
        ResidualUnit(256, rate=rates[1], name='4e'),
        ResidualUnit(256, rate=rates[1], name='4f'),

        # When stride=2 reduces width and hight by factor of 2
        ResidualUnit(512, rate=rates[1], name='5a',
                     stride=strides[2], has_branch=True),
        ResidualUnit(512, rate=rates[2], name='5b'),
        ResidualUnit(512, rate=rates[2], name='5c'),
    )

    if include_global_pool:
        resnet = layers.join(
            resnet,
            # Since the final residual unit has 2048 output filters, global
            # pooling will replace every output image with single average
            # value. Despite input image size, output from this layer always
            # will be a vector with 2048 values.
            layers.GlobalPooling('avg'),
            layers.Softmax(1000, name='fc1000'),
        )

    return resnet



import abc

import six
import numpy as np
import tensorflow as tf

from neupy.core.docs import SharedDocsABCMeta


__all__ = ('Initializer', 'Constant', 'Normal', 'Uniform', 'Orthogonal',
           'HeNormal', 'HeUniform', 'XavierNormal', 'XavierUniform')


def identify_fans(shape):
    """
    Identify fans from shape.

    Parameters
    ----------
    shape : tuple or list

    Returns
    -------
    tuple
        Tuple that contains :math:`fan_{in}` and :math:`fan_{out}`.
    """
    n_dimensions = len(shape)

    if n_dimensions == 0:
        raise ValueError("Cannot apply initializer when shape is unknown")

    elif n_dimensions == 1:
        fan_in, fan_out = shape[0], 1

    elif n_dimensions == 2:
        fan_in, fan_out = shape

    else:
        # By default we assume that weights with more than 2 dimensions
        # are generated for convolutional layers.
        receptive_field = np.prod(shape[:-2]).item(0)
        fan_in = shape[-2] * receptive_field
        fan_out = shape[-1] * receptive_field

    return fan_in, fan_out


def classname(instance):
    return instance.__class__.__name__


def set_numpy_seed(seed=None):
    if seed is not None:
        np.random.seed(seed)


class Initializer(six.with_metaclass(SharedDocsABCMeta)):
    """
    Base class for parameter initialization.

    Methods
    -------
    sample(shape, return_array=False)
        Returns tensorflow's tensor or numpy array with specified
        shape. Type of the object depends on the ``return_array`` value.
        Numpy array will be returned when ``return_array=True`` and
        tensor otherwise.
    """
    inherit_method_docs = True

    @abc.abstractmethod
    def sample(self, shape, return_array=False):
        """
        Returns tensorflow's tensor with specified shape.

        Parameters
        ----------
        shape : tuple
            Parameter shape.

        return_array : bool
            Returns numpy's array when equal to ``True``
            and tensorflow's tensor when equal to ``False``.
            Defaults to ``False``.

        Returns
        -------
        array-like or Tensor
        """
        raise NotImplementedError


class Constant(Initializer):
    """
    Initialize parameter that has constant values.

    Parameters
    ----------
    value : float, int
        All parameters in the tensor will be equal to
        this value. Defaults to ``0``.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, value=0):
        self.value = value

    def sample(self, shape, return_array=False):
        if return_array:
            return np.ones(shape) * self.value

        elif self.value == 0:
            return tf.zeros(shape)

        return tf.ones(shape) * self.value

    def __repr__(self):
        return '{}({})'.format(classname(self), self.value)


class Normal(Initializer):
    """
    Initialize parameter sampling from the normal distribution.

    Parameters
    ----------
    mean : int, float
        Mean of the normal distribution.

    std : int, float
        Standard deviation of the normal distribution.

    seed : None or int
        Random seed. Integer value will make results reproducible.
        Defaults to ``None``.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, mean=0, std=0.01, seed=None):
        self.mean = mean
        self.std = std
        self.seed = seed

    def sample(self, shape, return_array=False):
        if return_array:
            set_numpy_seed(self.seed)
            return np.random.normal(loc=self.mean, scale=self.std, size=shape)

        return tf.random_normal(
            mean=self.mean, stddev=self.std,
            shape=shape, seed=self.seed)

    def __repr__(self):
        return '{}(mean={}, std={})'.format(
            classname(self), self.mean, self.std)


class Uniform(Initializer):
    """
    Initialize parameter sampling from the uniform
    distribution.

    Parameters
    ----------
    minval : int, float
        Minimum possible value.

    maxval : int, float
        Maximum possible value.

    seed : None or int
        Random seed. Integer value will make results reproducible.
        Defaults to ``None``.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, minval=0, maxval=1, seed=None):
        self.minval = minval
        self.maxval = maxval
        self.seed = seed

    def sample(self, shape, return_array=False):
        minval, maxval = self.minval, self.maxval

        if return_array:
            set_numpy_seed(self.seed)
            return np.random.random(shape) * (maxval - minval) + minval

        return tf.random_uniform(shape, minval, maxval, seed=self.seed)

    def __repr__(self):
        return '{}({}, {})'.format(
            classname(self), self.minval, self.maxval)


class Orthogonal(Initializer):
    """
    Initialize matrix with orthogonal basis.

    Parameters
    ----------
    scale : int, float
        Scales output matrix by a specified factor.
        Defaults to ``1``.

    seed : None or int
        Random seed. Integer value will make results reproducible.
        Defaults to ``None``.

    Raises
    ------
    ValueError
        In case if tensor shape has more than 2
        dimensions.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, scale=1.0, seed=None):
        self.scale = scale
        self.seed = seed

    def sample(self, shape, return_array=False):
        ndim = len(shape)

        if ndim not in (1, 2):
            raise ValueError(
                "Shape attribute must have 1 or 2 dimensions. "
                "Found {} dimensions".format(ndim))

        normal = Normal(seed=self.seed)
        rand_matrix = normal.sample(shape, return_array)

        if ndim == 1:
            return rand_matrix

        if return_array:
            u, _, v = np.linalg.svd(rand_matrix, full_matrices=False)
        else:
            _, u, v = tf.linalg.svd(rand_matrix, full_matrices=False)
            v = tf.transpose(v)

        nrows, ncols = shape
        ortho_base = u if nrows > ncols else v
        return self.scale * ortho_base[:nrows, :ncols]

    def __repr__(self):
        return '{}(scale={})'.format(classname(self), self.scale)


class InitializerWithGain(Initializer):
    """
    Initialization class that has gain property

    Parameters
    ----------
    gain : float
        Scales variance of the distribution by this factor. Value ``2``
        is a suitable choice for layers that have Relu non-linearity.
        Defaults to ``1``.

    seed : None or int
        Random seed. Integer value will make results reproducible.
        Defaults to ``None``.
    """
    def __init__(self, gain=1.0, seed=None):
        self.gain = gain
        self.seed = seed
        super(InitializerWithGain, self).__init__()

    def __repr__(self):
        return '{}(gain={})'.format(classname(self), self.gain)


class HeNormal(InitializerWithGain):
    """
    Kaiming He parameter initialization method based on the
    normal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Kaiming He, Xiangyu Zhan, Shaoqing Ren, Jian Sun.
        Delving Deep into Rectifiers: Surpassing Human-Level
        Performance on ImageNet Classification, 2015.
    """
    def sample(self, shape, return_array=False):
        fan_in, _ = identify_fans(shape)
        variance = 1. / fan_in
        std = np.sqrt(self.gain * variance)

        normal = Normal(0, std, seed=self.seed)
        return normal.sample(shape, return_array)


class HeUniform(InitializerWithGain):
    """
    Kaiming He parameter initialization method based on the
    uniformal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Kaiming He, Xiangyu Zhan, Shaoqing Ren, Jian Sun.
        Delving Deep into Rectifiers: Surpassing Human-Level
        Performance on ImageNet Classification, 2015.
    """
    def sample(self, shape, return_array=False):
        fan_in, _ = identify_fans(shape)
        variance = 3. / fan_in
        abs_max_value = np.sqrt(self.gain * variance)

        uniform = Uniform(
            minval=-abs_max_value,
            maxval=abs_max_value,
            seed=self.seed,
        )
        return uniform.sample(shape, return_array)


class XavierNormal(InitializerWithGain):
    """
    Xavier Glorot parameter initialization method based on
    normal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Xavier Glorot, Y Bengio. Understanding the difficulty
        of training deep feedforward neural networks, 2010.
    """
    def sample(self, shape, return_array=False):
        fan_in, fan_out = identify_fans(shape)
        variance = 1. / (fan_in + fan_out)
        std = np.sqrt(self.gain * variance)

        normal = Normal(0, std, seed=self.seed)
        return normal.sample(shape, return_array)


class XavierUniform(InitializerWithGain):
    """
    Xavier Glorot parameter initialization method based
    on uniform distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Xavier Glorot, Y Bengio. Understanding the difficulty
        of training deep feedforward neural networks, 2010.
    """
    def sample(self, shape, return_array=False):
        fan_in, fan_out = identify_fans(shape)
        variance = 3. / (fan_in + fan_out)
        abs_max_value = np.sqrt(self.gain * variance)

        uniform = Uniform(
            minval=-abs_max_value,
            maxval=abs_max_value,
            seed=self.seed,
        )
        return uniform.sample(shape, return_array)



from __future__ import division

import numpy as np

from neupy import init
from neupy.utils import format_data
from neupy.exceptions import NotTrained
from neupy.algorithms.base import BaseNetwork
from neupy.core.properties import (
    IntProperty, Property,
    TypedListProperty, NumberProperty,
)


__all__ = ('LVQ', 'LVQ2', 'LVQ21', 'LVQ3')


def euclid_distance(X, weight):
    X = np.expand_dims(X, axis=0)
    euclid_dist = np.linalg.norm(X - weight, axis=1)
    return np.expand_dims(euclid_dist, axis=0)


def n_argmin(array, n, axis=0):
    sorted_argumets = array.argsort(axis=axis).ravel()
    return sorted_argumets[:n]


class LVQ(BaseNetwork):
    """
    Learning Vector Quantization (LVQ) algorithm.

    Notes
    -----
    - Input data needs to be normalized, because LVQ uses
      Euclidean distance to find clusters.

    - Training error is just a ratio of misclassified
      samples

    Parameters
    ----------
    n_inputs : int
        Number of input units. It should be equal to the
        number of features in the input data set.

    n_subclasses : int, None
        Defines total number of subclasses. Values should be greater
        or equal to the number of classes. ``None`` will set up number
        of subclasses equal to the number of classes. Defaults to ``None``
        (or the same as ``n_classes``).

    n_classes : int
        Number of classes in the data set.

    prototypes_per_class : list, None
        Defines number of prototypes per each class. For instance,
        if ``n_classes=3`` and ``n_subclasses=8`` then there are
        can be 3 subclasses for the first class, 3 for the second one
        and 2 for the third one (3 + 3 + 2 == 8). The following example
        can be specified as ``prototypes_per_class=[3, 3, 2]``.

        There are two rules that apply to this parameter:

        1. ``sum(prototypes_per_class) == n_subclasses``

        2. ``len(prototypes_per_class) == n_classes``

        The ``None`` value will distribute approximately equal
        number of subclasses per each class. It's approximately,
        because, for cases, when ``n_subclasses % n_classes != 0``
        there is no way to distribute equal number of subclasses
        per each class.

        Defaults to ``None``.

    {BaseNetwork.step}

    n_updates_to_stepdrop : int or None
        If this options is not equal to ``None`` then after every
        update LVQ reduces step size and do it until number of
        applied updates would reach the ``n_updates_to_stepdrop``
        value. The minimum possible step size defined in the
        ``minstep`` parameter.

        Be aware that number of updates is not the same as number
        of epochs. LVQ applies update after each propagated sample
        through the network. Relations between this parameter and
        maximum number of epochs is following

        .. code-block:: python

            n_updates_to_stepdrop = n_samples * n_max_epochs

        If parameter equal to ``None`` then step size wouldn't be
        reduced after each update.

        Defaults to ``None``.

    minstep : float
        Step size would never be lower than this value. This
        property useful only in case if ``n_updates_to_stepdrop``
        is not ``None``. Defaults to ``1e-5``.

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.signals}

    {Verbose.verbose}

    Methods
    -------
    {BaseSkeleton.predict}

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> X = np.array([[0, 0], [0, 1], [1, 0], [1, 1], [2, 2], [1, 2]])
    >>> y = np.array([0, 0, 0, 1, 1, 1])
    >>>
    >>> lvqnet = algorithms.LVQ(n_inputs=2, n_classes=2)
    >>> lvqnet.train(X, y, epochs=100)
    >>> lvqnet.predict([[2, 1], [-1, -1]])
    array([1, 0])
    """
    n_inputs = IntProperty(minval=1)
    n_subclasses = IntProperty(minval=2, default=None, allow_none=True)
    n_classes = IntProperty(minval=2)

    prototypes_per_class = TypedListProperty(allow_none=True, default=None)
    weight = Property(
        expected_type=(np.ndarray, init.Initializer),
        allow_none=True,
        default=None,
    )
    n_updates_to_stepdrop = IntProperty(
        default=None,
        allow_none=True,
        minval=1,
    )
    minstep = NumberProperty(minval=0, default=1e-5)

    def __init__(self, **options):
        self.initialized = False
        super(LVQ, self).__init__(**options)

        self.n_updates = 0

        if self.n_subclasses is None:
            self.n_subclasses = self.n_classes

        if isinstance(self.weight, init.Initializer):
            weight_shape = (self.n_inputs, self.n_subclasses)
            self.weight = self.weight.sample(weight_shape, return_array=True)

        if self.weight is not None:
            self.initialized = True

        if self.n_subclasses < self.n_classes:
            raise ValueError("Number of subclasses should be greater "
                             "or equal to the number of classes. Network "
                             "was defined with {} subclasses and {} classes"
                             "".format(self.n_subclasses, self.n_classes))

        if self.prototypes_per_class is None:
            whole, reminder = divmod(self.n_subclasses, self.n_classes)
            self.prototypes_per_class = [whole] * self.n_classes

            if reminder:
                # Since we have reminder left, it means that we cannot
                # have an equal number of subclasses per each class,
                # therefor we will add +1 to randomly selected class.
                class_indices = np.random.choice(self.n_classes, reminder,
                                                 replace=False)

                for class_index in class_indices:
                    self.prototypes_per_class[class_index] += 1

        if len(self.prototypes_per_class) != self.n_classes:
            raise ValueError("LVQ defined for classification problem that has "
                             "{} classes, but the `prototypes_per_class` "
                             "variable has defined data for {} classes."
                             "".format(self.n_classes,
                                       len(self.prototypes_per_class)))

        if sum(self.prototypes_per_class) != self.n_subclasses:
            raise ValueError("Invalid distribution of subclasses for the "
                             "`prototypes_per_class` variable. Got total "
                             "of {} subclasses ({}) instead of {} expected"
                             "".format(sum(self.prototypes_per_class),
                                       self.prototypes_per_class,
                                       self.n_subclasses))

        self.subclass_to_class = []
        for class_id, n_prototypes in enumerate(self.prototypes_per_class):
            self.subclass_to_class.extend([class_id] * n_prototypes)

    @property
    def training_step(self):
        if self.n_updates_to_stepdrop is None:
            return self.step

        updates_ratio = (1 - self.n_updates / self.n_updates_to_stepdrop)
        return self.minstep + (self.step - self.minstep) * updates_ratio

    def predict(self, X):
        if not self.initialized:
            raise NotTrained("LVQ network hasn't been trained yet")

        X = format_data(X)
        subclass_to_class = self.subclass_to_class
        weight = self.weight

        predictions = []
        for input_row in X:
            output = euclid_distance(input_row, weight)
            winner_subclass = int(output.argmin(axis=1))

            predicted_class = subclass_to_class[winner_subclass]
            predictions.append(predicted_class)

        return np.array(predictions)

    def train(self, X_train, y_train, *args, **kwargs):
        X_train = format_data(X_train)
        y_train = format_data(y_train)

        n_input_samples = len(X_train)

        if n_input_samples <= self.n_subclasses:
            raise ValueError("Number of training input samples should be "
                             "greater than number of subclasses. Training "
                             "method received {} input samples."
                             "".format(n_input_samples))

        if not self.initialized:
            target_classes = sorted(np.unique(y_train).astype(np.int))
            expected_classes = list(range(self.n_classes))

            if target_classes != expected_classes:
                raise ValueError(
                    "All classes should be integers from the range [0, {}], "
                    "but got the following classes instead {}"
                    "".format(self.n_classes - 1, target_classes))

            weights = []
            iterator = zip(target_classes, self.prototypes_per_class)
            for target_class, n_prototypes in iterator:
                is_valid_class = (y_train[:, 0] == target_class)
                is_valid_class = is_valid_class.astype('float64')
                n_samples_per_class = sum(is_valid_class)
                is_valid_class /= n_samples_per_class

                if n_samples_per_class <= n_prototypes:
                    raise ValueError(
                        "Input data has {0} samples for class-{1}. Number "
                        "of samples per specified class-{1} should be "
                        "greater than {2}.".format(
                            n_samples_per_class, target_class, n_prototypes))

                class_weight_indices = np.random.choice(
                    np.arange(n_input_samples), n_prototypes,
                    replace=False, p=is_valid_class)

                class_weight = X_train[class_weight_indices]
                weights.extend(class_weight)

            self.weight = np.array(weights)
            self.initialized = True

        super(LVQ, self).train(X_train, y_train, *args, **kwargs)

    def one_training_update(self, X_train, y_train):
        weight = self.weight
        subclass_to_class = self.subclass_to_class

        n_correct_predictions = 0
        for input_row, target in zip(X_train, y_train):
            step = self.training_step
            output = euclid_distance(input_row, weight)
            winner_subclass = int(output.argmin())
            predicted_class = subclass_to_class[winner_subclass]

            weight_update = input_row - weight[winner_subclass, :]
            is_correct_prediction = (predicted_class == target).item(0)

            if is_correct_prediction:
                weight[winner_subclass, :] += step * weight_update
            else:
                weight[winner_subclass, :] -= step * weight_update

            n_correct_predictions += is_correct_prediction
            self.n_updates += 1

        n_samples = len(X_train)
        return 1 - n_correct_predictions / n_samples


class LVQ2(LVQ):
    """
    Learning Vector Quantization 2 (LVQ2) algorithm.
    Improved version for the LVQ algorithm.

    Parameters
    ----------
    epsilon : float
        Ration between to closest subclasses that
        triggers double weight update. Defaults to ``0.1``.

    {LVQ.Parameters}

    Notes
    -----
    {LVQ.Notes}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> X = np.array([[0, 0], [0, 1], [1, 0], [1, 1], [2, 2], [1, 2]])
    >>> y = np.array([0, 0, 0, 1, 1, 1])
    >>>
    >>> lvqnet = algorithms.LVQ2(n_inputs=2, n_classes=2)
    >>> lvqnet.train(X, y, epochs=100)
    >>> lvqnet.predict([[2, 1], [-1, -1]])
    array([1, 0])
    """
    epsilon = NumberProperty(default=0.1)

    def one_training_update(self, X_train, y_train):
        weight = self.weight
        epsilon = self.epsilon
        subclass_to_class = self.subclass_to_class

        n_correct_predictions = 0
        for input_row, target in zip(X_train, y_train):
            step = self.training_step
            output = euclid_distance(input_row, weight)
            winner_subclasses = n_argmin(output, n=2, axis=1)

            top1_subclass, top2_subclass = winner_subclasses
            top1_class = subclass_to_class[top1_subclass]
            top2_class = subclass_to_class[top2_subclass]

            top1_weight_update = input_row - weight[top1_subclass, :]
            is_correct_prediction = (top1_class == target).item(0)

            closest_dist, runner_up_dist = output[0, winner_subclasses]
            double_update_condition_satisfied = (
                not is_correct_prediction and
                (top2_class == target) and
                closest_dist > ((1 - epsilon) * runner_up_dist) and
                runner_up_dist < ((1 + epsilon) * closest_dist)
            )

            if double_update_condition_satisfied:
                top2_weight_update = input_row - weight[top2_class, :]
                weight[top1_subclass, :] -= step * top1_weight_update
                weight[top2_subclass, :] += step * top2_weight_update

            elif is_correct_prediction:
                weight[top1_subclass, :] += step * top1_weight_update

            else:
                weight[top1_subclass, :] -= step * top1_weight_update

            n_correct_predictions += is_correct_prediction

        n_samples = len(X_train)
        return 1 - n_correct_predictions / n_samples


class LVQ21(LVQ2):
    """
    Learning Vector Quantization 2.1 (LVQ2.1) algorithm.
    Improved version for the LVQ2 algorithm.

    Parameters
    ----------
    {LVQ2.Parameters}

    Notes
    -----
    {LVQ2.Notes}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> X = np.array([[0, 0], [0, 1], [1, 0], [1, 1], [2, 2], [1, 2]])
    >>> y = np.array([0, 0, 0, 1, 1, 1])
    >>>
    >>> lvqnet = algorithms.LVQ21(n_inputs=2, n_classes=2)
    >>> lvqnet.train(X, y, epochs=100)
    >>> lvqnet.predict([[2, 1], [-1, -1]])
    array([1, 0])
    """
    def one_training_update(self, X_train, y_train):
        weight = self.weight
        epsilon = self.epsilon
        subclass_to_class = self.subclass_to_class

        n_correct_predictions = 0
        for input_row, target in zip(X_train, y_train):
            step = self.training_step
            output = euclid_distance(input_row, weight)
            winner_subclasses = n_argmin(output, n=2, axis=1)

            top1_subclass, top2_subclass = winner_subclasses
            top1_class = subclass_to_class[top1_subclass]
            top2_class = subclass_to_class[top2_subclass]

            top1_weight_update = input_row - weight[top1_subclass, :]
            is_correct_prediction = (top1_class == target).item(0)

            closest_dist, runner_up_dist = output[0, winner_subclasses]
            double_update_condition_satisfied = (
                (
                    (top1_class == target and top2_class != target) or
                    (top1_class != target and top2_class == target)
                ) and
                closest_dist > ((1 - epsilon) * runner_up_dist) and
                runner_up_dist < ((1 + epsilon) * closest_dist)
            )

            if double_update_condition_satisfied:
                top2_weight_update = input_row - weight[top2_class, :]

                if is_correct_prediction:
                    weight[top2_subclass, :] -= step * top2_weight_update
                    weight[top1_subclass, :] += step * top1_weight_update
                else:
                    weight[top1_subclass, :] -= step * top1_weight_update
                    weight[top2_subclass, :] += step * top2_weight_update

            elif is_correct_prediction:
                weight[top1_subclass, :] += step * top1_weight_update

            else:
                weight[top1_subclass, :] -= step * top1_weight_update

            n_correct_predictions += is_correct_prediction
            self.n_updates += 1

        n_samples = len(X_train)
        return 1 - n_correct_predictions / n_samples


class LVQ3(LVQ21):
    """
    Learning Vector Quantization 3 (LVQ3) algorithm.
    Improved version for the LVQ2.1 algorithm.

    Parameters
    ----------
    {LVQ.n_inputs}

    {LVQ.n_subclasses}

    {LVQ.n_classes}

    {LVQ.prototypes_per_class}

    {LVQ2.epsilon}

    slowdown_rate : float
        Paremeter scales learning step in order to decrease it
        in case if the two closest subclasses predict target
        value correctly. Defaults to ``0.4``.

    step : float
        Learning rate, defaults to ``0.01``.

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.signals}

    {Verbose.verbose}

    Notes
    -----
    {LVQ21.Notes}
    - Decreasing step and increasing number of training epochs
      can improve the performance.

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> X = np.array([[0, 0], [0, 1], [1, 0], [1, 1], [2, 2], [1, 2]])
    >>> y = np.array([0, 0, 0, 1, 1, 1])
    >>>
    >>> lvqnet = algorithms.LVQ3(n_inputs=2, n_classes=2)
    >>> lvqnet.train(X, y, epochs=100)
    >>> lvqnet.predict([[2, 1], [-1, -1]])
    array([1, 0])
    """
    step = NumberProperty(minval=0, default=0.01)
    slowdown_rate = NumberProperty(minval=0, default=0.4)

    def one_training_update(self, X_train, y_train):
        weight = self.weight
        epsilon = self.epsilon
        slowdown_rate = self.slowdown_rate
        subclass_to_class = self.subclass_to_class

        n_correct_predictions = 0
        for input_row, target in zip(X_train, y_train):
            step = self.training_step
            output = euclid_distance(input_row, weight)
            winner_subclasses = n_argmin(output, n=2, axis=1)

            top1_subclass, top2_subclass = winner_subclasses
            top1_class = subclass_to_class[top1_subclass]
            top2_class = subclass_to_class[top2_subclass]

            top1_weight_update = input_row - weight[top1_subclass, :]
            is_first_correct = (top1_class == target).item(0)
            is_second_correct = (top2_class == target).item()

            closest_dist, runner_up_dist = output[0, winner_subclasses]
            double_update_condition_satisfied = (
                (
                    (is_first_correct and not is_second_correct) or
                    (is_second_correct and not is_first_correct)
                ) and
                closest_dist > ((1 - epsilon) * runner_up_dist) and
                runner_up_dist < ((1 + epsilon) * closest_dist)
            )
            two_closest_correct_condition_satisfied = (
                is_first_correct and is_second_correct and
                closest_dist > ((1 - epsilon) * (1 + epsilon) * runner_up_dist)
            )

            if double_update_condition_satisfied:
                top2_weight_update = input_row - weight[top2_class, :]

                if is_first_correct:
                    weight[top1_subclass, :] += step * top1_weight_update
                    weight[top2_subclass, :] -= step * top2_weight_update
                else:
                    weight[top1_subclass, :] -= step * top1_weight_update
                    weight[top2_subclass, :] += step * top2_weight_update

            elif two_closest_correct_condition_satisfied:
                beta = step * slowdown_rate
                top2_weight_update = input_row - weight[top2_class, :]

                weight[top1_subclass, :] += beta * top1_weight_update
                weight[top2_subclass, :] += beta * top2_weight_update

            else:
                weight[top1_subclass, :] -= step * top1_weight_update

            n_correct_predictions += is_first_correct
            self.n_updates += 1

        n_samples = len(X_train)
        return 1 - n_correct_predictions / n_samples



__all__ = (
    'LayerConnectionError', 'InvalidConnection', 'NotTrained',
    'StopTraining', 'WeightInitializationError', 'PropagationError',
)


class LayerConnectionError(Exception):
    """
    Error class that triggers in case of connection
    issues within layers.
    """


class InvalidConnection(Exception):
    """
    Connection is not suitable for the specified algorithm
    """


class NotTrained(Exception):
    """
    Algorithms hasn't been trained yet.
    """


class StopTraining(Exception):
    """
    Interrupt training procedure.
    """


class WeightInitializationError(Exception):
    """
    In case if there is some problem with weight initialization
    """


class PropagationError(Exception):
    """
    Error propagation triggers when input cannot be propagated
    through the layer.
    """
